#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <pthread.h>
#include <time.h>

#define N 10 // broj delova takodje i broj threadova
#define X 3 // kapacitet susare
#define S 2 // duzina susenja

pthread_mutex_t mutex[2];
pthread_cond_t condFarbar; // uslov za farbare
pthread_cond_t condSusar; // uslov za susaru

int povrsine[N]; // niz povrsina za svaki deo
int slojevi[N]; // niz slojeva koje je potrebno ofarbati za svaki deo

int naSusenju = 0; // trenutno delova u susari
int naFarbanju1 = 0; // broj delova na farbanju kod farbara1
int naFarbanju2 = 0; // broj delova na farbanju kod farbara2

void *workFarbanje(void *arg)
{
	long id = (long) arg;
	int brojSlojeva = 0; // broj vec ofarbanih slojeva za deo
	int duzinaFarbanja = 0;
	int kojiFarbar = 0; // 1 farbar1, 2 farbar2
	 
	int t = rand()%5; // posle t trenutaka kada deo dodje kod farbara
	sleep(t);
		
	while(1)
	{
		//FARBANJE
		
		pthread_mutex_lock(mutex+0); // zakljucava farbare
		while(naFarbanju1 == 1 && naFarbanju2 == 1) // ako su obojica zauzeta cekaj da se javi neki slobodan
			pthread_cond_wait(&condFarbar, mutex+0);
		
		duzinaFarbanja = povrsine[id];
		if(naFarbanju1 == 0) // ako je slobodan farbar1
		{
			//printf("FARBAR-1 : DEO-%ld\n", id);
			kojiFarbar = 1;
			naFarbanju1 = 1;
		}	
		else if(naFarbanju2 == 0) // ako je slobodan farbar2
		{
			//printf("FARBAR-2 : DEO-%ld\n", id);
			kojiFarbar = 2;
			naFarbanju2 = 1;
		}
		
		pthread_mutex_unlock(mutex+0); // otkljucava farbare
		
		brojSlojeva++; // nanosi jedan sloj boje
		printf("Farbar%d farba %d sloj %ld dela \n", kojiFarbar, brojSlojeva, id);
		sleep(duzinaFarbanja); // farba deo Pi sekundi 
		
		pthread_mutex_lock(mutex+0);
		
			if(kojiFarbar == 1) // ako si bio kod farbara1 oslobodi mu naFarbanju1
			{
				naFarbanju1 = 0; // slobodan je farbar1 za sledeceg
				pthread_cond_broadcast(&condFarbar); // javljamo svima koji cekaju farbara
			}
			else if(kojiFarbar == 2) // ako si bio kod farbara2 oslobodi mu naFarbanju2
			{
				naFarbanju2 = 0; // slobodan je farbar2 za sledeceg
				pthread_cond_broadcast(&condFarbar); // javljamo svima koji cekaju farbara
			}
			kojiFarbar = 0; // resetujemo na 0 kako bi posle dobili prvog slobodnog
		
		pthread_mutex_unlock(mutex+0);
		
		//SUSENJE
		
		pthread_mutex_lock(mutex+1); // kad je nanet sloj boje mora da ide deo kod farbara
			while(naSusenju >= X) // sve dok je susara puna cekaj da se oslobodi mesto
				pthread_cond_wait(&condSusar, mutex+1); // cekam na signal da je susara slobodna
			
			naSusenju++; // povecavam broj delova u susari
		pthread_mutex_unlock(mutex+1); // otkljucavam susaru
			
			printf("Deo %ld se susi u susari %d put.\n", id, brojSlojeva);
			sleep(S); // svaki deo se susi S sekundi 
				
		pthread_mutex_lock(mutex+1);
			naSusenju--; // deo je zavrsio sa susenjem
			pthread_cond_broadcast(&condSusar); // javljam svima koji cekaju susenje da mogu da udju
			if(brojSlojeva == slojevi[id])// ako je zavrsio sa susenjem zadnjeg sloja
			{
				pthread_mutex_unlock(mutex+1);
				printf("Susara zavrsila DEO - %ld\n", id);
				pthread_exit(NULL);
			} 
		pthread_mutex_unlock(mutex+1);
	}	
}

int main()
{
	long i;
	pthread_t threads[N];
	
	for(i=0; i<N; i++)
	{
		povrsine[i] = rand()%3 + 1; // rand povrsina Pi od 1-3
		slojevi[i] = rand()%3 + 1; // rand slojeva potrebnih Ki od 1-3
	}
	
	for(i=0;i<2;i++)
		pthread_mutex_init(mutex+i, NULL); //0 semafor za farbare, 1 semafor za susara

	pthread_cond_init(&condFarbar, NULL); // uslov za farbare
	pthread_cond_init(&condSusar, NULL); // uslov za susara
	
	for(i=0; i<N; i++)
	{
		pthread_create(threads+0, NULL, workFarbanje, (void*)i); // svaki deo poslat na farbanje
	}
	
	for(i=0;i<N;i++)
		pthread_join(threads[i], NULL);
	
	return 0;
}
