#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <unistd.h>
#include <math.h>

// dva radnika i susara su deljeni resursi
// deo dodje na farbanje, posle svakog sloja ide u susaru

// susara ima kapacitet X
// dakle u jednom trenutku 2 radnika i susara sa X kapacitetom

#define BROJ_DELOVA 200000
#define DUZINA_SUSENJA 2

int slobodniRadnici = 2;
int slobodnoMestaSusara = 5;

pthread_mutex_t mutexFarbanje;
pthread_mutex_t mutexSusenje;

pthread_cond_t condFarbanje;
pthread_cond_t condSusenje;


void* obradaDela(void* arg)
{
	long id = (long) arg;
	sleep(rand() % 2);
	int brojSlojeva = (rand() % 4) + 1;
	int duzinaFarbanja = (rand() % 2) + 1;
	
	printf("Ja sam deo %ld i stigao sam u farbaru. Broj slojeva: %d.\n", id, brojSlojeva);
	while(brojSlojeva > 0)
	{		
		// pokusaj farbanja
		pthread_mutex_lock(&mutexFarbanje);
		while(slobodniRadnici == 0)
			pthread_cond_wait(&condFarbanje, &mutexFarbanje);
		slobodniRadnici--;
		pthread_mutex_unlock(&mutexFarbanje);
		
		sleep(duzinaFarbanja);
		brojSlojeva--; // farbanje
		printf("Deo %ld: Ofarban sloj: %d.\n", id, brojSlojeva+1);
		
		pthread_mutex_lock(&mutexFarbanje);
		slobodniRadnici++;
		pthread_mutex_unlock(&mutexFarbanje);
		pthread_cond_signal(&condFarbanje);
		
		// pokusaj susenja
		pthread_mutex_lock(&mutexSusenje);
		while(slobodnoMestaSusara == 0)
			pthread_cond_wait(&condSusenje, &mutexSusenje);
		slobodnoMestaSusara--; // susenje
		pthread_mutex_unlock(&mutexSusenje);
		
		sleep(DUZINA_SUSENJA);
		printf("Deo %ld: Osusen sloj: %d.\n", id, brojSlojeva+1);
		
		pthread_mutex_lock(&mutexSusenje);
		slobodnoMestaSusara++;
		pthread_mutex_unlock(&mutexSusenje); 
		pthread_cond_signal(&condSusenje);
		
		printf("Deo %ld: Preostalo slojeva: %d.\n", id, brojSlojeva);
	}
		
	printf("Ja sam deo %ld i zavrsio sam.\n", id);
	pthread_exit(NULL);
}

int main(int argc, char **argv)
{
	pthread_t t[BROJ_DELOVA];
	pthread_mutex_init(&mutexFarbanje, NULL);
	pthread_mutex_init(&mutexSusenje, NULL);
	
	pthread_cond_init(&condFarbanje, NULL);
	pthread_cond_init(&condSusenje, NULL);
	
	for(long i=0;i<BROJ_DELOVA;i++)
		pthread_create(&t[i], NULL, obradaDela, (void*) i);
	
	for(int i=0;i<BROJ_DELOVA;i++)
		pthread_join(t[i], NULL);
		
	pthread_mutex_destroy(&mutexFarbanje);
	pthread_mutex_destroy(&mutexSusenje);
		
	pthread_cond_destroy(&condFarbanje);
	pthread_cond_destroy(&condSusenje);
	
	return EXIT_SUCCESS;
}
