#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdbool.h>
#include <pthread.h>

#define VREME_FARBANJA 1
#define VREME_SUSENJA 3
#define KAPACITET_SUSARE 5
#define BROJ_DELOVA 20

bool farbar1Zauzet = false;
bool farbar2Zauzet = false;

pthread_mutex_t farbar1Lock;
pthread_mutex_t farbar2Lock;
pthread_mutex_t mestaUSusari[KAPACITET_SUSARE];

typedef struct deo{
    int deoID;
    int vremeDolaska;
    int povrsina;
    int brSlojeva;
}Deo;

Deo* initDeo(int deoID, int vremeDolaska, int povrsina, int brSlojeva){
    Deo *new = malloc(sizeof(Deo));

    new->deoID = deoID;
    new->vremeDolaska = vremeDolaska;
    new->povrsina = povrsina;
    new->brSlojeva = brSlojeva;

    return new;
}

void* obradi(void* deoData){

    pthread_mutex_t daLiSmeDaSeSusi;
    pthread_mutex_init(&daLiSmeDaSeSusi, NULL);
    int brojSlojeva = 0;

    Deo *deo = (Deo*) deoData;

    sleep(deo->vremeDolaska);

    for(int i = 0 ; i < deo->brSlojeva ; i++){

        pthread_mutex_lock(&daLiSmeDaSeSusi);

        int zauzetFarbar = 0;

        if(pthread_mutex_trylock(&farbar1Lock) == 0){
            
            zauzetFarbar = 1;
            farbar1Zauzet = true;
            printf("FARBAR 1: Farbam %d sloj dela %d\n", i, deo->deoID);
            sleep(VREME_FARBANJA * deo->povrsina);
            pthread_mutex_unlock(&daLiSmeDaSeSusi);
        }
        else{
            if(pthread_mutex_trylock(&farbar2Lock) == 0){
                
                zauzetFarbar = 2;
                farbar2Zauzet = true;
                printf("FARBAR 2: Farbam %d sloj dela %d\n", i, deo->deoID);
                sleep(VREME_FARBANJA * deo->povrsina);
                pthread_mutex_unlock(&daLiSmeDaSeSusi);
            }
        }



        if(zauzetFarbar == 1){
            farbar1Zauzet = false;
            pthread_mutex_unlock(&farbar1Lock);
        }
        else if(zauzetFarbar == 2){
            farbar2Zauzet = false;
            pthread_mutex_unlock(&farbar2Lock);
        }


        if(pthread_mutex_trylock(&daLiSmeDaSeSusi) != 0){
            i--;
            pthread_mutex_unlock(&daLiSmeDaSeSusi);
            continue;
        }

        for(int j = 0 ; j < KAPACITET_SUSARE ; j++){
            if(pthread_mutex_trylock(&mestaUSusari[j]) == 0)
            {   
                printf("MESTO %d je zauzeto za susenje sledecih %d sekundi, susim deo %d\n", j, VREME_SUSENJA, deo->deoID);
                sleep(VREME_SUSENJA);
                pthread_mutex_unlock(&mestaUSusari[j]);
                break;
            }
        }

        pthread_mutex_unlock(&daLiSmeDaSeSusi);

    }

}

int main(){

    pthread_t delovi[BROJ_DELOVA];

    pthread_mutex_init(&farbar1Lock, NULL);
    pthread_mutex_init(&farbar2Lock, NULL);

    for(int i = 0; i < KAPACITET_SUSARE; i++){
        pthread_mutex_init(&mestaUSusari[i], NULL);
    }


    for(int i = 0 ; i < BROJ_DELOVA ; i++){
        
        int vremeDolaska = rand() % 10 + 1;
        int povrsina = rand() % 5 + 1;
        int brSlojeva = rand() % 2 + 1;

        Deo *deo = initDeo(i, vremeDolaska, povrsina, brSlojeva);

        pthread_create(&delovi[i], NULL, &obradi, (void*) deo);
    }

    for(int i = 0 ; i < BROJ_DELOVA ; i++){
        pthread_join(delovi[i], NULL);
    }

    return 0;
}