/* Krunoslav Velfel, 49/2018 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>

struct dio_ {
    int pi; /* povrsina */
    int ki; /* br. slojeva farbe */
    int s;  /* vrijeme u susari */
};

/* X: najveci dozvoljeni broj dijelova u susari u istom trenutku (definisati po potrebi) */
const int X = 3;

/* Mutexi: d - niz delova koje je potrebno obraditi, f - farbara, s - susara */
pthread_mutex_t d, f, s;

/* Prebrojavanje dijelova u susari: */

const int MAX_RADNIKA = 2;
const int MAX_DIJELOVA = 100;

pthread_cond_t susara;
struct dio_ ss[MAX_DIJELOVA];
int cs = 0;

pthread_cond_t farbara;
struct dio_ ff[MAX_DIJELOVA];
int cf = 0;

/* Trenutno definisano 8 dijelova sa razlicitim parametrima (definisati po potrebi)*/
struct dio_ dio [] = {
    {.pi = 1, .ki = 1, .s = 10},
    {.pi = 1, .ki = 1, .s = 10},
    {.pi = 1, .ki = 1, .s = 10},
    {.pi = 1, .ki = 1, .s = 10},
    {.pi = 5, .ki = 1, .s = 10},
    {.pi = 5, .ki = 1, .s = 10},
    {.pi = 5, .ki = 1, .s = 10},
    {.pi = 5, .ki = 1, .s = 10}};

/* Pokreni onoliko niti koliko ima i dijelova u gornjem nizu, koje treba ofarbati: */
const int THREADS = sizeof(dio) / sizeof(struct dio_);


/* Proces obrade svakog od dijelova: */
void *farbar_work(void *arg) {
    int id = *((int*)arg);

    for(;;) {
        struct dio_ dd;
        pthread_mutex_lock(&f);
        while(cf <= 0) {
            printf("[FARBAR id = %d] Nema dijelova za farbanje\n", id);
            pthread_cond_wait(&farbara, &f);
        }
        dd = ff[cf];
        cf--;
        printf("[FARBAR id = %d] Uzeo sam dio, farbam %d sekundi\n", id, dd.pi);
        sleep(dd.pi);
        dd.ki--;
        pthread_mutex_unlock(&f);

        /* Ofarbani dio proslijedi susari */
        pthread_mutex_lock(&s);
        ss[cs] = dd;
        cs++;
        pthread_mutex_unlock(&s);
        pthread_cond_broadcast(&susara);
    }

    return NULL;
}

void *susara_work(void *arg) {
    int id = *((int*)arg);

    for(;;) {
        struct dio_ dd;
        pthread_mutex_lock(&s);
        while(cs <= 0) {
            printf("[SUSAR id = %d] Nema dijelova za susenje\n", id);
            pthread_cond_wait(&susara, &s);
        }
        dd = ss[cs];
        cs--;
        printf("[SUSAR id = %d] Uzeo sam dio, susim %d sekundi\n", id, dd.s);
        sleep(dd.pi);
        pthread_mutex_unlock(&s);

        /* Osuseni dio proslijedi nazad farbari, ako je potrebno */
        if(dd.ki > 0) {
            pthread_mutex_lock(&f);
            ff[cf] = dd;
            cf++;
            pthread_mutex_unlock(&f);
            pthread_cond_broadcast(&farbara);
        }

    }

    return NULL;
}

/* Proces obrade svakog od dijelova: */
void *work(void *arg) {
    int id = *((int*)arg);

    pthread_mutex_lock(&d);
    struct dio_ tmp = dio[id];
    int pi = dio[id].pi;
    int ki = dio[id].ki;
    int ss = dio[id].s;
    pthread_mutex_unlock(&d);

    printf("[DEO i = %d] Stizem u farbaru\n", id);

    pthread_mutex_lock(&f);
    ff[cf] = tmp;
    cf++;
    pthread_mutex_unlock(&f);
    pthread_cond_broadcast(&farbara);
 /*
    for(int i = 0; i < ki; i++) {
        pthread_mutex_lock(&f);
        printf("[DEO i = %d] Farbar nanosi jedan sloj farbe povrsine pi = %d\n", id, pi);
        sleep(pi);
        pthread_mutex_unlock(&f);

        printf("[DEO i = %d] Farbar je zavrsio posao, idem u susaru\n", id);

        pthread_mutex_lock(&s);
        while(count_susara >= X) {
            printf("[DEO i = %d] Susara je ZAUZETA, cekam...\n", id);
            pthread_cond_wait(&susara, &s);
        }
        count_susara++;
        pthread_mutex_unlock(&s);
        
        printf("[DEO i = %d] U susari sam sljedecih s = %d sekundi\n", id, ss);
        sleep(ss);

        pthread_mutex_lock(&s);
        count_susara--;
        pthread_mutex_unlock(&s);
        pthread_cond_broadcast(&susara);
    }

    printf("[DEO i = %d] Zavrsio sam, izlazim iz farbare\n", id);*/
    return NULL;
}

int main(void) {
    pthread_t threads[2 * MAX_RADNIKA + THREADS];
    int thread_args[2 * MAX_RADNIKA + THREADS];

    pthread_cond_init(&susara, NULL);
    pthread_cond_init(&farbara, NULL);
    
    pthread_mutex_init(&d, NULL);
    pthread_mutex_init(&f, NULL);
    pthread_mutex_init(&s, NULL);

        int i = 0;

    for(i = 0; i < THREADS; i++) {
        thread_args[i] = i;
        pthread_create(&threads[i], NULL, work, &thread_args[i]);
    }

        for(; i < 2 * MAX_RADNIKA + THREADS; i++) {
        thread_args[i] = i;
        pthread_create(&threads[i], NULL, farbar_work, &thread_args[i]);
        pthread_create(&threads[i], NULL, susara_work, &thread_args[i]);
    }

    printf("[MAIN THREAD] Sve niti su pokrenute\n");

    for(int i = 0; i < 2 * MAX_RADNIKA + THREADS; i++) {
        pthread_join(threads[i], NULL);
    }

    printf("[MAIN THREAD] Svi dijelovi su obradjeni\n");

    pthread_cond_destroy(&susara);
    pthread_cond_destroy(&farbara);
    
    pthread_mutex_destroy(&d);
    pthread_mutex_destroy(&f);
    pthread_mutex_destroy(&s);

    return EXIT_SUCCESS;
}