#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>
#include <time.h>
#include <stdbool.h>

#define SLOBODAN -1
#define PRAZNO -1

#define BROJ_DELOVA 5
#define BROJ_MESTA 2
#define BROJ_RADNIKA 2

#define MAX_P 5
#define MAX_K 3
#define MAX_T 20
#define OFFSET_T 3
#define S 5

typedef struct deo 
{
    int id;
    int t;
    int p;
    int k;
} Deo;

int *mestaUSusari;
int *radnici;
int trenutnoSeFarba;
int trenutnoSeSusi;

pthread_mutex_t mutex;
pthread_cond_t condFarbanje;
pthread_cond_t condSusenje;

int dajSlobodnogRadnika()
{
    for (int i = 0; i < BROJ_RADNIKA; i++)
    {
        if (radnici[i] == SLOBODAN)
        {
            return i;
        }
    }
    
    return false;
}

int dajSlobodnoMesto()
{
    for (int i = 0; i < BROJ_MESTA; i++)
    {
        if (mestaUSusari[i] == SLOBODAN)
        {
            return i;
        }
    }
    
    return false;
}

void* tracker(void* argument)
{
    while (1)
    {
        sleep(1);

        printf("susara (%d/%d), R0: %d, R1: %d\n", trenutnoSeSusi, BROJ_MESTA, radnici[0], radnici[1]);
    }
}

void* farbanje(void* argument)
{
    int radnik, mesto, sloj; 
    Deo *deo = (Deo*)argument;

    sleep(deo->t);

    printf("Deo %d: stigao sam (P: %d, K: %d), cekam farbanje\n", deo->id, deo->p, deo->k);

    for (int i = 1; i <= deo->k; i++)
    {
        sloj = i;
        
        // CEKAM RED ZA FARBANJE
        pthread_mutex_lock(&mutex);
        {
            while (!(trenutnoSeFarba < BROJ_RADNIKA))
            {
                pthread_cond_wait(&condFarbanje, &mutex);
            }

            // RADNIK SE OSLOBODIO
            trenutnoSeFarba++;
            radnik = dajSlobodnogRadnika();
            radnici[radnik] = deo->id;
        }
        pthread_mutex_unlock(&mutex);

        pthread_cond_broadcast(&condFarbanje);

        // FARBAM SE
        printf("Deo %d: radnik %d pocinje da farba %d. sloj\n", deo->id, radnik, sloj);
        
        sleep(deo->p);

        pthread_mutex_lock(&mutex);
        {
            // OFARBAN SAM, OSLOBADJAM RADNIKA
            radnici[radnik] = SLOBODAN;
            trenutnoSeFarba--;
        }
        pthread_mutex_unlock(&mutex);

        printf("Deo %d: radnik %d je zavrsio da farba %d. sloj\n", deo->id, radnik, sloj);

        printf("Deo %d: cekam susenje %d. sloja\n", deo->id, sloj);

        // CEKAM SUSENJE
        pthread_mutex_lock(&mutex);
        {
            while (!(trenutnoSeSusi < BROJ_MESTA))
            {
                pthread_cond_wait(&condSusenje, &mutex);
            }

            // MESTO U SUSARI SE OSLOBODILO
            trenutnoSeSusi++;
            mesto = dajSlobodnoMesto();
            mestaUSusari[mesto] = deo->id;
        }
        pthread_mutex_unlock(&mutex);

        pthread_cond_broadcast(&condSusenje);

        // SUSIM SE
        printf("Deo %d: pocinjem sa susenjem %d. sloja\n", deo->id, sloj);
        
        sleep(S);
        
        pthread_mutex_lock(&mutex);
        {
            // OSUSIO SAM SE, OSLOBADJAM MESTO U SUSARI
            mestaUSusari[mesto] = PRAZNO;
            trenutnoSeSusi--;
        }
        pthread_mutex_unlock(&mutex);

        printf("Deo %d: osusio se %d. sloj\n", deo->id, sloj);
    }

    printf("Deo %d: zavrsio sam sa kompletnim farbanjem\n", deo->id);
}

int main()
{
    pthread_t *threads;

    trenutnoSeFarba = 0;
    trenutnoSeSusi = 0;

    mestaUSusari = (int*)malloc(BROJ_MESTA * sizeof(int));

    for (int i = 0; i < BROJ_MESTA; i++)
    {
        mestaUSusari[i] = PRAZNO;
    }

    radnici = (int*)malloc(BROJ_RADNIKA * sizeof(int));

    for (int i = 0; i < BROJ_RADNIKA; i++)
    {
        radnici[i] = SLOBODAN;
    }

    // TRACKER (fixan za 2 radnika, kod radi i sa vise od 2 radnika)
    pthread_t trackerThread;

    pthread_create(&trackerThread, NULL, tracker, NULL);
    // END TRACKER

    pthread_mutex_init(&mutex, NULL);

    pthread_cond_init(&condFarbanje, NULL);
    pthread_cond_init(&condSusenje, NULL);

    threads = (pthread_t*)malloc(BROJ_DELOVA * sizeof(pthread_t));

    for (int i = 0; i < BROJ_DELOVA; i++)
    {
        Deo *deo = (Deo*)malloc(sizeof(Deo));

        deo->id = i;
        deo->t = rand() % MAX_T + OFFSET_T;
        deo->p = rand() % MAX_P + 1;
        deo->k = rand() % MAX_K + 1;

        pthread_create(threads + i, NULL, farbanje, (void*)deo);
    }

    for (int i = 0; i < BROJ_DELOVA; i++)
    {
        pthread_join(threads[i], NULL);
    }

    return 0;
}