/*
    t - vreme kada stize u farbaru
    p - povrsina dela (kolika je povrsina toliko sekundi se farba jedan sloj)
    Samo jednom sloju je potrebno p sekundi da se ofarba, pa se on osusi, pa se farba ponovo
    k - koliko je slojeva farbe potrebno naneti na deo
    s - koliko sekundi se susi deo
    x - koliko delova susara moze da susi u jednom trenutku

    Dva radnika koji nezavisno vrse farbanje i susara predstavljaju deljene resurse i svi
    delovi im pristupaju ravnopravno
*/

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <unistd.h>
#include <time.h>

#define BROJ_DELOVA 10
#define BROJ_SLOJEVA 4
#define MAX_POVRSINA 4

#define VREME_SUSENJA 2
#define KAPACITET_SUSARE 1

#define BROJ_FARBARA 2

pthread_mutex_t mutex_farbanje, mutex_susenje;
pthread_cond_t cond_farbanje, cond_susenje;

// Deljeni resursi
int br_zauzetih_farbara = 0;
int br_delova_u_susari = 0;

void *deo(void *arg)
{
    long id = (long)arg;

    int povrsina_dela = (rand() % MAX_POVRSINA) + 1;

    sleep(rand() % 2);
    printf("DEO %ld, POVRSINA %d: Stigao na farbanje!\n", id+1, povrsina_dela);

    int broj_ofarbanih_slojeva = 0;

    while(broj_ofarbanih_slojeva < BROJ_SLOJEVA)
    {
        // Farbanje
        pthread_mutex_lock(&mutex_farbanje);
        while(br_zauzetih_farbara == BROJ_FARBARA) pthread_cond_wait(&cond_farbanje, &mutex_farbanje);

        br_zauzetih_farbara++;

        pthread_mutex_unlock(&mutex_farbanje);

        
        printf("DEO %ld, POVRSINA %d: Farbanje %d. sloja...\n", id+1, povrsina_dela, broj_ofarbanih_slojeva+1);
        sleep(povrsina_dela);

        
        pthread_mutex_lock(&mutex_farbanje);

        br_zauzetih_farbara--;
        pthread_cond_broadcast(&cond_farbanje);
        //pthread_cond_signal(&cond_farbanje);
        
        pthread_mutex_unlock(&mutex_farbanje);

        // Susenje
        pthread_mutex_lock(&mutex_susenje);
        while(br_delova_u_susari == KAPACITET_SUSARE) pthread_cond_wait(&cond_susenje, &mutex_susenje);

        br_delova_u_susari++;

        pthread_mutex_unlock(&mutex_susenje);

        
        printf("DEO %ld, POVRSINA %d: Susenje %d. sloja...\n", id+1, povrsina_dela, broj_ofarbanih_slojeva+1);
        sleep(VREME_SUSENJA);

        
        pthread_mutex_lock(&mutex_susenje);
        
        br_delova_u_susari--;
        pthread_cond_broadcast(&cond_susenje);
        //pthread_cond_signal(&cond_susenje);

        pthread_mutex_unlock(&mutex_susenje);

        broj_ofarbanih_slojeva++;
    }

    printf("DEO %ld, POVRSINA %d: Uspesno ofarban!\n", id+1, povrsina_dela);
}


int main()
{
    pthread_t delovi[BROJ_DELOVA];

    pthread_mutex_init(&mutex_farbanje, NULL);
    pthread_mutex_init(&mutex_susenje, NULL);

    pthread_cond_init(&cond_farbanje, NULL);
    pthread_cond_init(&cond_susenje, NULL);

    for(long i = 0; i < BROJ_DELOVA; i++) pthread_create(delovi + i, NULL, deo, (void*)i);
    for(long i = 0; i < BROJ_DELOVA; i++) pthread_join(delovi[i], NULL);

    return 0;
}